/*
 * civil_untils.c
 *
 *  Created on: Jul 13, 2009
 *      Author: yanming
 */

/**
 * Convert recombination frequency to Haldane map distance.
 *
 * @param r a combination frequency
 *
 * @return d Haldane map distance
 */
double rho2dis (double r)
{
   return -log (1.0 - 2.0 * r) / 2.0 * 100;
}

/**
 * Convert Haldane map distance to recombination frequency
 *
 * @param d: Halane map distance
 *
 * @return recombination frequency
 *
 */
double dis2rho (double d)
{
   return 0.5 * (1.0 - exp (-2.0 * (d / 100)));
}

/**
`* Compute the number of 1's in a bitwise representation of an
 * integer i for i=0, ..., 2^la.
 */
void compute_ca (char *ca, int la)
{
   /* ca[i] stores the number of 1's in a bitwise representation of an
      integer i, i=0, ..., 2^la. */

   int   j;
   unsigned long i;
   unsigned long n0, n1;

   n0 = 1;
   ca[0] = 0;
   for (j = 0; j < la; j++)
   {
      memcpy (&ca[n0], ca, n0);
      n1 = n0 << 1;
      for (i = n0; i < n1; i++)
      {
         ca[i]++;
      }
      n0 = n1;
   }
}

/**
 * Simulate a binary (inheritance) vector.
 *
 * In this implementation, only one random number needs to be generated.
 *
 * @param n the length of the vector
 * @param p a vector of mean values
 * @return the simulated inheritance vector
 */
/*inline*/
iv_t simulate_binary_vec (int n, double *p)
{
   double p_prefix, p_threshold;
   double x;
   int   i;
   iv_t  s;

   x = rans ();
   s = 0;
   p_prefix = 1;
   p_threshold = 1 - p[0];

   for (i = 0; i < n - 1; i++)
   {
      if (x > p_threshold)
      {
         s += 1 << i;
         p_prefix *= p[i];
         p_threshold += p_prefix * (1 - p[i + 1]);
      }
      else
      {
         p_prefix *= (1 - p[i]);
         p_threshold -= p_prefix * p[i + 1];
      }
   }

   if (x > p_threshold)
   {
      s += 1 << (n - 1);
   }

   return s;
}

/**
 * Map an inheritance vector to the representative element in the same
 * equivalence class.
 *
 * In a pedigree, due to founder phase uncertainty, probabilities (and
 * w-scores, and Bayes scores, etc.) of two inheritance vectors will be the
 * same if the two inheritance vectors differ only in founder phase.
 *
 * For each founder in the pedigree, all his/her children will have a meiosis
 * indicator corresponding to this founder. We put these meiosis indicators
 * corresponding to a same founder parent in one group. In each group, we can
 * hold one meiosis indicator fixed and not affect the calculations of
 * likelihood, w-score, Bayes scores etc. In other words, in an inheritance
 * vector, if we flip ALL meiosis indicators in a meiosis group defined as
 * above, we will get an equivalent inheritance vector.
 *
 * To uniquely identify a representative inheritance vector in each
 * equivalence class, we can fix the meiosis indicator of the first child of
 * each founder parent to be 0.
 *
 * @param s an inheritance vector
 *
 * @param adg_pt contains the following fields:
 *
 *  --n_fixed_bits number of bits (meiosis indicators) that can be fixed
 *
 *  --fixed_bit_masks indicate bits that can be fixed.
 *
 *  --mei_grp_masks each entry indicates all bits within a meiosis group.
 *
 * @return the representative element of the equivalent class that s belongs to.
 */
/* inline */
iv_t rep_iv (iv_t s, const adg_t * adg_pt)
{
   int   i;                             /* indexes meioses */
   for (i = 0; i < adg_pt->n_fixed_bits; i++)
   {
      if (s & adg_pt->fixed_bit_masks[i])
      {
         s ^= adg_pt->mei_grp_masks[i];
      }
   }
   return s;
}

/**
 * Suppress the fixed bits of an inheritance vector.
 *
 * @param s a (full) inheritance vector
 *
 * @param adg_pt contains the following fields:
 *
 *   -- n_non_fixed_bits: number of non-fixed bits
 *
 *   -- bit_masks: indicate which bits in the full inheritance vector are
 *   non-fixed.
 */
/* inline */
iv_t reduced_iv (iv_t s, const adg_t * adg_pt)
{
   iv_t  red_s = 0;                     /* Reduced inheritance vector. */
   int   i;

   for (i = 0; i < adg_pt->n_non_fixed_bits; i++)
   {
      if (s & adg_pt->non_fixed_bit_masks[i])
         red_s += 1 << i;
   }

   return red_s;
}

/**
 * Expand a reduced inheritance vector to full length by inserting 0's at
 * those fixed bits.
 *
 * @param red_s a reduced inheritance vector, where fixed bits were suppressed
 *
 * @param adg_pt contains the following fields:
 *
 *   -- n_non_fixed_bits: number of non-fixed bits
 *
 *   -- bit_masks: indicate which bits in the full inheritance vector are
 *   non-fixed.
 */
/* inline */
iv_t full_iv (iv_t red_s, adg_t * adg_pt)
{
   iv_t  s = 0;                         /* Reduced inheritance vector. */
   int   i;

   for (i = 0; i < adg_pt->n_non_fixed_bits; i++)
   {
      if (red_s & (1 << i))
         s += adg_pt->non_fixed_bit_masks[i];
   }

   return s;
}

/**
 * Compute a table that maps full ivs to reduced ivs.
 *
 * @param adg_pt: points to a pointer to an allele descent graph.
 *
 */
void compute_iv_map (const adg_t * adg_pt, iv_t ** red_s_ppt)
{
   iv_t  n_ivs, full_s, rep_s;
   n_ivs = 1 << adg_pt->n_nfa;

   *red_s_ppt = malloc (sizeof (iv_t) * n_ivs);
   for (full_s = 0; full_s < n_ivs; full_s++)
   {
      rep_s = rep_iv (full_s, adg_pt);
      (*red_s_ppt)[full_s] = reduced_iv (rep_s, adg_pt);
   }
}

/**
 * Derive founder genome labels based on the inheritance vector.
 *
 *  @param s an inheritance vector
 *
 *  @param adg_pt: points to an allelic descent graph containing the following
 *  fields
 *
 *     n_fa: number of founder alleles
 *
 *     i_fa[]: indices to founder alleles
 *
 *     n_nfa: number of non-founder alleles
 *
 *     i_nfa[]: indices to non-founder alleles.
 *
 *     i_maa[]: indices to parent's paternal allele.
 *
 *     i_paa[]: indices to parent's maternal allele.
 *
 * @param fgl will store the founder genome labels
 *
 * Notes:
 *
 *   . The indices are of alleles, not of individuals. This is more flexible.
 *   For example, it will allow us to deal with individuals with only one
 *   parent in the pedigree.
 *
 *   . Each non-founder allele corresponds to one meiosis, thus one bit of s.
 *
 *   . If the i's bit of s is 0, then allele i_nfa[i] and allele i_maa[i]
 *   should have the same fgl.  if the i's bit of S and 1, then allele
 *   i_nfa[i] and allele i_paa[i] should have the same fgl.
 *
 *   . In i_nfa, we require parent alleles appear before corresponding children
 *   alleles.
 *
 *   . length of fgl should be n_fa + n_nfa;
 */
void iv2fgl (const iv_t s, const adg_t * adg_pt, int *fgl)
{

   int   i;
   int   i_a;                           /* index to current allele */

   for (i = 0; i < adg_pt->n_fa; i++)
   {
      fgl[adg_pt->i_fa[i]] = i;
   }

   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      i_a = adg_pt->i_nfa[i];
      if ((s >> i) & 1)
      {
         fgl[i_a] = fgl[adg_pt->i_paa[i]];
      }
      else
      {
         fgl[i_a] = fgl[adg_pt->i_maa[i]];
      }
      /* fgl[i_a] = fgl[i_maa[i] - ((s>>i) & 1)]; */
   }
}

/**
 * Reduce and relabel the fgl vector.
 *
 * @param full_fgl fgl of all individuals in a pedigree
 *
 * @param adg_pt points to an allele descent graph, which contains the following fields,
 *
 *     n_obs: number of individuals with observed trait values
 *
 *     i_obsa: indicies of alleles of observed individuals, length(i_obsa) = n_obs * 2.
 *
 * @param red_fgl (output)fgl of individuals with observed trait values, relabeled to 0,
 *   1, ...,  max_fgl.
 *
 */
int reduce_relabel_fgl (int *full_fgl, adg_t * adg_pt, int *red_fgl)
{
   int   i;                             /*indexes meioses */
   int   n_obs2;
   int   max_fgl;
   int   label;
   int  *map;

   n_obs2 = adg_pt->n_obs * 2;

   /* The largest possible foudner allele label is n_fa! */
   /* NOT n_obs * 2! */
   map = (int *) malloc (adg_pt->n_fa * sizeof (int));

   max_fgl = 0;
   for (i = 0; i < adg_pt->n_fa; i++)
   {
      map[i] = -1;
   }

   for (i = 0; i < n_obs2; i++)
   {
      label = full_fgl[adg_pt->i_obsa[i]];
      if (map[label] < 0)
      {
         map[label] = max_fgl;
         max_fgl++;
      }
      red_fgl[i] = map[label];
   }

   free (map);
   return max_fgl;
}

/**
 * Simulate inheritance vectors.
 *
 * @param adg_list_pt points to a list of pedigrees
 *
 * @param iv_list_pt points to corresponding inheritance vectors
 */
void simulate_rep_ivs (const adg_list_t * adg_list_pt, iv_list_t * iv_list_pt)
{
   iv_t  s_max;
   int   h;                             /*indexes samples */
   int   j;                             /*indexes loci */
   int   k;                             /*indexes pedigree */

   iv_list_pt->n_peds = adg_list_pt->n_peds;

   iv_list_pt->s = (iv_t ***) malloc (sizeof (iv_t **) * iv_list_pt->smpl_size);
   for (h = 0; h < iv_list_pt->smpl_size; h++)
   {
      iv_list_pt->s[h] =
         (iv_t **) malloc (sizeof (iv_t *) * iv_list_pt->n_loci);
      for (j = 0; j < iv_list_pt->n_loci; j++)
      {
         iv_list_pt->s[h][j] =
            (iv_t *) malloc (sizeof (iv_t) * iv_list_pt->n_peds);
      }
   }

   j = 0;
   for (k = 0; k < adg_list_pt->n_peds; k++)
   {
      /* s_max = 1<<adg_list_pt->peds[k].n_nfa-1; */
      s_max = (1 << adg_list_pt->peds[k].n_non_fixed_bits) - 1;

      for (h = 0; h < iv_list_pt->smpl_size; h++)
      {
         /* sample s unifromly fro 0 to s_max */
         iv_list_pt->s[h][j][k] = intran (0, s_max);
      }
   }

   /* Need to simulate linked S to simulate S at a linked locus */
}
