
/**
 * Compute the z vector, z'z and sum z (for a single pedigree): z is the vector of cross-products of mean-adjusted trait values.
 * These parts are needed to compute a locus-specific additive variance  component score vc(fgl, y).
  *
 * @param y a vector of observed quantitative trait values
 *
 * @param n_obs length of the vector y (number of observed trait values)
 *
 * @param mu  estimated mean trait value (e.g. estimated from a large population)
 *
 * @param sd estimated sd of the trait values (e.g.estiamted from a large population). the specification of sd does not affect the value of the vc score.
 *
 * @param z (return value) the vector of trait value cross-product
 *
 * @param sumz (return value) sum z
 *
 * @param zz (return value) z'z.
 *
 */
void vc_z (double *y, int n_obs, double mu, double sd, double *z, double *sumz,
           double *zz)
{
   int   i, i1, i2;
   double sd2;

   /* The vector of trait value products:
    *
    *  z = (y1y2, y1y3, ..., yn-1yn).
    */

   i = 0;
   *sumz = 0;
   *zz = 0;
   sd2 = sd * sd;
   for (i1 = 0; i1 < n_obs - 1; i1++)
   {
      for (i2 = (i1 + 1); i2 < n_obs; i2++)
      {
         z[i] = (y[i1] - mu) * (y[i2] - mu) / sd2;
         *sumz += z[i];
         *zz += z[i] * z[i];
         i++;
      }
   }
}

/**
 * Compute sum phi, phi'z phi'phi for all representative inheritance vectors on a pedigree.
 *
 *
 * @param fgl a vector of length 2 x nobs x n_fgl, each (2 x nobs) subvector
 * is one assignment of FGL's to observed individuals in the pedigree.
 *
 * @param y a vector of quantitative trait values.
 *
 * @param n_obs number of observed individuals (length of y).
 *
 * @param sumphi (return value) sum phi, a length n_fgl vector
 *
 * @param phiz (return value) phi'z, a length n_fgl vector
 *
 * @param phiphi (return value) phi'phi, a length n_fgl vector
 */
void vc_parts (int *fgl, int n_obs, double *z, int *sumphi, double *phiz,
               int *phiphi)
{

   int   i, i1, i2;
   int   ig, a0, a1, b0, b1, phi;

   /* Compute phi'z and phi'phi for each fgl vector. */
   *sumphi = 0;
   *phiphi = 0;
   *phiz = 0.0;

   i = 0;
   for (i1 = 0; i1 < n_obs - 1; i1++)
   {
      ig = i1 * 2;
      a0 = fgl[ig++];
      a1 = fgl[ig++];

      for (i2 = i1 + 1; i2 < n_obs; i2++)
      {
         b0 = fgl[ig++];
         b1 = fgl[ig++];
         phi = (a0 == b0) + (a0 == b1) + (a1 == b0) + (a1 == b1);
#ifdef DEBUG
         printf
            ("(i1, i2): (%d, %d), fgl: (%d, %d) (%d, %d), phi: %d, (yi, yj)=(%4.1lf, %4.1lf), (y_i-my)*(y_j-my)=%6.1lf \n",
             i1, ig / 2, a0, a1, b0, b1, phi, y[i1], y[i2], z[i]);
#endif
         if (phi > 0)
         {
            *sumphi += phi;
            *phiz += z[i] * phi;
            *phiphi += phi * phi;
         }
         i++;
      }
   }
}

/**
 * Compute the parts needed to compute a vc score and store them to files.
 *
 * Let z be the vector of cross-products of mean-adjusted trait values
 *
 *   z = ((y1-mu)(y2-mu), ..., (y_n-1 -mu)(yn-mu))
 *
 * and phi the vector of IBD sharing
 *
 *   phi = (phi_12, ..., phi_n-1,n),
 *
 * where (phi_i1, i2) is the number of alleles shared between individuals i1, i2.
 *
 * If we fit a linear regression model,
 *
 *   z = a + b phi + e
 *
 * the vc-score is defined to be the t-test statistic for the regression coef. b.
 *
 * Equivalently, let r be the correlation coefficient between z and phi:
 *
 *   r = (phi'z - sum z sum phi/n) / sqrt((phi'phi - sum phi^2/n)(z'z - sum z^2/n))
 *
 * The vc score can be computed as
 *
 *       vc = r / sqrt(1 - r^2)
 *
 * Two inheritance vectors are equivalent if they differ only by
 * founder phase. Accordingly, we can partition the set of all possible
 * inheritance vectors on each pedigree into equivalence classes. We will
 * choose one element from each equivalence class and compute for all
 * these representative inheritance vectors the parts needed to compute the
 * vc scores. We will store the computed values into files named after the
 * corresponding names of the pedigrees.
 *
 * In the output file, we will store the length of the z-vector, z'z value
 * for this pedigree, sum(z), sum(phi), phi'z for each fgl and phi'phi for each fgl.
 *
 * @param adg_list_t a list of allele descent graphs. An allele descent graph
 * is one abstract way to represent a pedigree.
 *
 * @param dir the directory where the computed values will be stored.
 *
 */
void compute_vc_tables (adg_list_t * adg_pt, trait_model_t * trt_mod_pt,
                        char *dir)
{
   FILE *file_out;
   file_name_t file_name;

   iv_t  s, red_s;
   int   k;                             /* indexes pedigrees. */
   long  n_rep_s;

   int   fgl[128];                      /* Potentail bug! */
   int   red_fgl[128];
   int   max_fgl;

   int   nz;
   double *z, zz, sumz, *phiz;
   int  *phiphi, *sumphi;

   /* k = 0; */
   for (k = 0; k < adg_pt->n_peds; k++)
   {
      /* name the output file */
      sprintf (file_name, "%s/vc/vc.%s.out", dir, adg_pt->peds[k].name);
      /* fprintf(stdmsg, "Processing pedigree %d ...\n", k); */
      /* fprintf(stdmsg, "%s\n", file_name); */
      /*
         fprintf(stdout,
         "adg graph: n_fa = %d, n_obs = %d, n_non_fixed_bits = %d\n",
         adg_pt->peds[k].n_fa, adg_pt->peds[k].n_obs,
         adg_pt->peds[k].n_non_fixed_bits);
       */

      file_out = fopen (file_name, "r"); /* check whether the file already exists */
      if (file_out == NULL)
      {                                 /* the file does not exist. */
         n_rep_s = 1 << adg_pt->peds[k].n_non_fixed_bits;
         /* number of possible inheritance vectors for this pedigree, module equivalence classes */

         nz = adg_pt->peds[k].n_obs * (adg_pt->peds[k].n_obs - 1) / 2;
         z = (double *) malloc (sizeof (double) * nz);
         vc_z (adg_pt->peds[k].y, adg_pt->peds[k].n_obs, trt_mod_pt->mu,
               trt_mod_pt->sd, z, &sumz, &zz);

         sumphi = (int *) malloc (sizeof (int) * n_rep_s);
         phiz = (double *) malloc (sizeof (double) * n_rep_s);
         phiphi = (int *) malloc (sizeof (int) * n_rep_s);

         for (red_s = 0; red_s < n_rep_s; red_s++)
         {
            s = full_iv (red_s, &adg_pt->peds[k]);

            /* print_iv(red_s, n_non_fixed_bits); */
            /* print_iv(s1, n_nfa); */
            /*   print_adg(&adg_pt->peds[k]); */

            iv2fgl (s, &adg_pt->peds[k], fgl);
            max_fgl = reduce_relabel_fgl (fgl, &adg_pt->peds[k], red_fgl);

            /* print_fgl(fgl, n_nfa); */
            /* print_fgl(red_fgl, n_obs*2 ); */

            vc_parts (red_fgl, adg_pt->peds[k].n_obs, z, &sumphi[red_s],
                      &phiz[red_s], &phiphi[red_s]);

         }

         file_out = fopen (file_name, "wb");
         /* file_out = fopen("scores.out", "w"); */
         if (file_out == NULL)
         {
            fprintf (stdmsg, "Cannot open file %s for writing\n", file_name);
            exit (1);
         }

         fwrite (&nz, sizeof (int), 1, file_out);
         fwrite (&sumz, sizeof (double), 1, file_out);
         fwrite (&zz, sizeof (double), 1, file_out);
         fwrite (sumphi, sizeof (int), n_rep_s, file_out);
         fwrite (phiz, sizeof (double), n_rep_s, file_out);
         fwrite (phiphi, sizeof (int), n_rep_s, file_out);
         fclose (file_out);
         fprintf (stdmsg, "Created %s\n", file_name);

         free (sumphi);
         free (phiz);
         free (phiphi);
         free (z);
      }
      else
      {
         fprintf (stdmsg, "will use existing table %s\n", file_name);
         fclose (file_out);
      }
   }
}

/**
 * Read pre-computed nz, z'z, phi'z, phi'phi from a file.
 *
 * @param adg_pt points to a list of pedigrees
 *
 * @param vc_tab_pt points to a structure that will be used to store nz, z'z, phi'z and phi'phi.
 *
 * @param dir where the computed values are stored
 *
 * @param k pedigree id. determines the name of the score file
 *
 */
void read_vc_table (const adg_list_t * adg_pt, vc_table_t * vc_tab_pt,
                    char *dir, int k)
{
   FILE *file_out;
   file_name_t file_name;

   vc_tab_pt->n_rep_s = 1 << adg_pt->peds[k].n_non_fixed_bits;
   vc_tab_pt->sumphi = (int *) malloc (sizeof (int) * vc_tab_pt->n_rep_s);
   vc_tab_pt->phiz = (double *) malloc (sizeof (double) * vc_tab_pt->n_rep_s);
   vc_tab_pt->phiphi = (int *) malloc (sizeof (int) * vc_tab_pt->n_rep_s);

   /* Name of the file that stores the precomputed values */
   sprintf (file_name, "%s/vc/vc.%s.out", dir, adg_pt->peds[k].name);
   file_out = fopen (file_name, "rb");
   if (file_out == NULL)
   {
      printf ("Can not open file %s for reading.\n", file_name);
      exit (1);
   }
   fread (&vc_tab_pt->nz, sizeof (int), 1, file_out);
   fread (&vc_tab_pt->sumz, sizeof (double), 1, file_out);
   fread (&vc_tab_pt->zz, sizeof (double), 1, file_out);
   fread (vc_tab_pt->sumphi, sizeof (int), vc_tab_pt->n_rep_s, file_out);
   fread (vc_tab_pt->phiz, sizeof (double), vc_tab_pt->n_rep_s, file_out);
   fread (vc_tab_pt->phiphi, sizeof (int), vc_tab_pt->n_rep_s, file_out);
   fclose (file_out);
}

/**
 * Print vc table.  One table for each pedigree. Each table stores the
 * w-scores computed for all representative inheritance vectors for one
 * pedigree.
 *
 * @param vc_tabs_pt points to the w-score tables.
 */
void print_vc_table (vc_table_t * pt)
{
   long  s;

   for (s = 0; s < pt->n_rep_s; s++)
   {
      printf ("%f, %4d\n", pt->phiz[s], pt->phiphi[s]);
   }
   printf ("\n");

}

/**
 * Marginal inheritance vector tests, using the variance component score as
 * test statistic.
 *
 * @param adg_pt points to a list to pedigrees.
 *
 * @param s_pt points to sampled inheritance vectors.
 *
 * @param pars_pt parameters needed for setting up the marginal tests
 *
 * @param dir the directory that store the vc tables.
 *
 */
void miv_vc (const adg_list_t * adg_pt, const iv_list_t * s_pt,
             miv_pars_t * pars_pt, char *dir)
{

   int   k;                             /*indexes pedigree */
   int   j;                             /*indexes loci */
   int   h, h1;                         /*indexes samples */
   int   n_sig;

   iv_t  s1;
   iv_t *red_s_pt;

   iv_list_t s0_pt;

   int   nz;
   double zz, sumz, muz, vz;
   int  *sumphi0, **sumphi;
   double *phiz0, **phiz;
   int  *phiphi0, **phiphi;

   double *t0, t, r;

   vc_table_t vc_tab;

   /* Simulate representative null s'es */
   printf ("Simulate representative null inheritance vectors.\n");
   s0_pt.smpl_size = pars_pt->n_mc;
   s0_pt.n_loci = 1;
   s0_pt.map = NULL;
   simulate_rep_ivs (adg_pt, &s0_pt);

   /* Compute vc socres */
   /* We will lookup nz, z'z, phi'z, phi'phi from pre-computed tables. */
   printf ("Compute z'z, phi'z, phi'phi.\n");

   /* Initialize variables */
   t0 = (double *) malloc (sizeof (double) * pars_pt->n_mc);

   sumphi0 = (int *) malloc (sizeof (int) * pars_pt->n_mc);
   phiz0 = (double *) malloc (sizeof (double) * pars_pt->n_mc);
   phiphi0 = (int *) malloc (sizeof (int) * pars_pt->n_mc);
   for (h = 0; h < pars_pt->n_mc; h++)
   {
      sumphi0[h] = 0;
      phiz0[h] = 0;
      phiphi0[h] = 0;
   }

   pars_pt->smpl_size = s_pt->smpl_size;
   pars_pt->t = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   pars_pt->p = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   sumphi = (int **) malloc (sizeof (int *) * pars_pt->smpl_size);
   phiz = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   phiphi = (int **) malloc (sizeof (int *) * pars_pt->smpl_size);

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      pars_pt->t[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      pars_pt->p[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      sumphi[h] = (int *) malloc (sizeof (int) * pars_pt->n_pos);
      phiz[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      phiphi[h] = (int *) malloc (sizeof (int) * pars_pt->n_pos);
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         pars_pt->t[h][j] = 0;          /* Why do I need this here? */
         sumphi[h][j] = 0;
         phiz[h][j] = 0;
         phiphi[h][j] = 0;
      }
   }

   nz = 0;
   zz = 0;
   sumz = 0;

   for (k = 0; k < adg_pt->n_peds; k++)
   {
      /* Compute peds[k].red_s here. */
      compute_iv_map (&adg_pt->peds[k], &red_s_pt);

      /* Read precomputed nz, z'z, phi'z, phi'phi. */
      read_vc_table (adg_pt, &vc_tab, dir, k);
      sumz += vc_tab.sumz;
      nz += vc_tab.nz;
      zz += vc_tab.zz;

      for (h = 0; h < pars_pt->n_mc; h++)
      {
         sumphi0[h] += vc_tab.sumphi[s0_pt.s[h][0][k]];
         phiz0[h] += vc_tab.phiz[s0_pt.s[h][0][k]];
         phiphi0[h] += vc_tab.phiphi[s0_pt.s[h][0][k]];
      }

      for (h = 0; h < s_pt->smpl_size; h++)
      {
         for (j = 0; j < pars_pt->n_pos; j++)
         {
            /* rep_s = rep_iv(s_pt->s[h][pars_pt->pos[j]][k], &adg_pt->peds[k]); */
            /* red_s = reduced_iv(rep_s, &adg_pt->peds[k]); */
            /* */
            /* Find the reprentative s corresponding to the observed s of
             * */
            /* pedigree k, at test position j, and in MCMC sample h. */
            s1 = red_s_pt[s_pt->s[h][pars_pt->pos[j]][k]];

            sumphi[h][j] += vc_tab.sumphi[s1];
            phiz[h][j] += vc_tab.phiz[s1];
            phiphi[h][j] += vc_tab.phiphi[s1];
         }
         /*     printf("\n"); */
      }

      free (red_s_pt);
      free_vc_table (&vc_tab);
   }
   free_iv_list (&s0_pt);

   printf ("Compute variance component score.\n");
   muz = sumz / nz;
   vz = (zz - sumz * sumz / nz);

   /* Compute null test statistics */
   for (h = 0; h < pars_pt->n_mc; h++)
   {
      /* t0[h] = phiz0[h] / sqrt((zz * phiphi0[h] - phiz0[h] * phiz0[h]) / nz); */
      /* print(t0[h]); */
      r = (phiz0[h] - sumphi0[h] * muz) / sqrt ((phiphi0[h]
                                                 -
                                                 sumphi0[h] * sumphi0[h] / nz) *
                                                vz);
      t0[h] = r * sqrt ((nz - 2) / (1 - r * r));
   }
   free (sumphi0);
   free (phiz0);
   free (phiphi0);

   /* Compute test statistic for the observed */
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         r = (phiz[h][j] - sumphi[h][j] * muz) / sqrt ((phiphi[h][j]
                                                        -
                                                        sumphi[h][j] *
                                                        sumphi[h][j] / nz) *
                                                       vz);
         pars_pt->t[h][j] = r * sqrt ((nz - 2) / (1 - r * r));
      }
      free (sumphi[h]);
      free (phiz[h]);
      free (phiphi[h]);
   }
   free (sumphi);
   free (phiz);
   free (phiphi);

   /*    printf("Compute p-values.\n"); */
   /*    for (h=0; h<s_pt->smpl_size; h++) { */
   /*        for (j=0; j<pars_pt->n_pos; j++) { */
   /*            printf("%8.5le, ", pars_pt->t[h][j]); */
   /*        } */
   /*        printf("\n"); */
   /*    } */
   /* */
   /*    for (h1=0; h1<pars_pt->n_mc; h1++) { */
   /*        printf("%8.5le, ", t0[h1]); */
   /*    }; */
   /*    printf("\n"); */

   printf ("Compute p-values.\n");
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         n_sig = 0;
         t = pars_pt->t[h][j];
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            n_sig += (t0[h1] > t);
         };
         pars_pt->p[h][j] = ((double) n_sig + 1) / ((double) pars_pt->n_mc + 1);
      }
   }

   free (t0);
}

/**
 * Conditional inheritance vector tests using variance component score as test
 * statistic.
 *
 * To compute the test statistic, we need to compte nz, z'z, phi'z and phi'phi.
 * For each single pedigree, we read these values from precomputed tables.
 * We then need sum over single-pedigree values to get the corresponding
 * multiple-pedigree values. The formula for computing the vc score is:
 *
 *       vc = r * sqrt((n-2) /(1 - r^2)
 *
 * where
 *
 *   r = (phi'z - sum z sum phi/n) / sqrt((phi'phi - sum phi^2/n)(z'z - sum z^2/n)).
 *
 * The challenge here is bookkeeping. For conditional test, at each locus, we
 * need to compute n.mcmc x n.mc test statistic. In an attempt to use each
 * vc-table (there is one for each pedigree) only once, we need to store
 *
 *   n.mcmc x n.mc x n.pos
 *
 * temporary sum(phi), phi'z, phi'phi (nz, sum z and z'z do not depend on phi, one copy
 * of each will be enough). This can take about 2G memory if there are 200 test
 * positions.
 *
 * @param adg_pt points to a list to pedigrees.
 *
 * @param s_pt points to sampled inheritance vectors.
 *
 * @param pars_pt parameters for setting the tests (test positions,
 * conditoning positions, etc.)
 *
 * @param dir working directory (will look for vc tables there).
 */
void civ_vc (const adg_list_t * adg_pt, const iv_list_t * s_pt, civ_pars_t
             * pars_pt, char *dir)
{

   int   i;                             /*indexes meioses */
   int   k;                             /*indexes pedigree */
   int   j;                             /*indexes loci */
   int   h, h1;                         /*indexes samples */
   int   n_sig;
   int   n_meis;

   iv_t  sL, sR;
   double rL, rR;
   iv_t  s0, s1;
   /*    iv_t rep_s, */
   iv_t *red_s_pt;

   double p0, p1, p_tab[2][2];          /* */
   double *p_sw;                        /* switching probabilities */

   int   bL, bR;

   int   nz;
   double zz, muz, vz;
   double sumz;

   int ***sumphi0;
   int ***phiphi0;
   /* double ***phiz0; */
   double ***t0;                        /*will first store phiz0, then later the null vc-scores. */

   int **sumphi;
   double **phiz;
   int **phiphi;
   double r, t;

   vc_table_t vc_tab;

   pars_pt->smpl_size = s_pt->smpl_size;

   fprintf (stdmsg, "Initialize variables for computing vc scores.\n");
   /* sumphi0, phiphi0, and t0 will take a huge amount of memory! */
   sumphi0 = (int ***) malloc (sizeof (int **) * pars_pt->smpl_size);
   phiphi0 = (int ***) malloc (sizeof (int **) * pars_pt->smpl_size);
   t0 = (double ***) malloc (sizeof (double **) * pars_pt->smpl_size);

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      sumphi0[h] = (int **) malloc (sizeof (int *) * pars_pt->n_pos);
      phiphi0[h] = (int **) malloc (sizeof (int *) * pars_pt->n_pos);
      /* phiz0[h]= (double **) malloc(sizeof(double *) * pars_pt->n_pos); */
      t0[h] = (double **) malloc (sizeof (double *) * pars_pt->n_pos);
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         sumphi0[h][j] = (int *) malloc (sizeof (int) * pars_pt->n_mc);
         phiphi0[h][j] = (int *) malloc (sizeof (int) * pars_pt->n_mc);
         t0[h][j] = (double *) malloc (sizeof (double) * pars_pt->n_mc);
         /* phiz0[h][j] = (double *) malloc(sizeof(double) * pars_pt->n_mc); */
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            sumphi0[h][j][h1] = 0;
            phiphi0[h][j][h1] = 0;
            /* phiz0[h][j][h1]=0; */
            t0[h][j][h1] = 0;
         }
      }
   }

   sumphi = (int **) malloc (sizeof (double *) * pars_pt->smpl_size);
   phiphi = (int **) malloc (sizeof (int *) * pars_pt->smpl_size);
   phiz = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   pars_pt->t = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   pars_pt->p = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      sumphi[h] = (int *) malloc (sizeof (double) * pars_pt->n_pos);
      phiz[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      phiphi[h] = (int *) malloc (sizeof (int) * pars_pt->n_pos);
      pars_pt->t[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      pars_pt->p[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         sumphi[h][j] = 0;
         phiz[h][j] = 0.0;
         phiphi[h][j] = 0;
      }
   }

   fprintf (stdmsg, "Compute vc scores.\n");

   nz = 0;
   zz = 0.0;
   sumz = 0.0;

   for (k = 0; k < adg_pt->n_peds; k++)
   {
      fprintf (stdmsg, "Processing pedigree %d ...\n", k);
      fflush (stdmsg);
      /* Read precomputed nz, z'z, phi'z, phi'phi. */
      read_vc_table (adg_pt, &vc_tab, dir, k);

      /* Compute a table that maps full ivs to reduced ivs. */
      compute_iv_map (&adg_pt->peds[k], &red_s_pt);

      nz += vc_tab.nz;
      sumz += vc_tab.sumz;
      zz += vc_tab.zz;

      n_meis = adg_pt->peds[k].n_nfa;
      p_sw = (double *) malloc (sizeof (double) * n_meis);

      /* Compute sum phi, phi'z and phi'phi for observed s. */
      for (h = 0; h < s_pt->smpl_size; h++)
      {
         for (j = 0; j < pars_pt->n_pos; j++)
         {
            /* printf("(j, j1, j2) = (%d, %d, %d)\n", */
            /*    pars_pt->pos[j], pars_pt->posl[j], pars_pt->posr[j]); */
            /*  rep_s = rep_iv(s_pt->s[h][pars_pt->pos[j]][k], &adg_pt->peds[k]); */
            /* red_s = reduced_iv(rep_s, &adg_pt->peds[k]); */
            /* */
            /* the following is equivalent to the above */
            s1 = red_s_pt[s_pt->s[h][pars_pt->pos[j]][k]];

            sumphi[h][j] += vc_tab.sumphi[s1];
            phiz[h][j] += vc_tab.phiz[s1];
            phiphi[h][j] += vc_tab.phiphi[s1];
         }
         /*     printf("\n"); */
      }

      /*    printf("(h, j, h1) = (%d, %d, %d)\n",  s_pt->smpl_size, pars_pt->n_pos, pars_pt->n_mc); */
      /* */

      /* Compute sum phi, phi'z and phi'phi for conditional null s0. */
      for (h = 0; h < s_pt->smpl_size; h++)
      {
         for (j = 0; j < pars_pt->n_pos; j++)
         {
            sL = s_pt->s[h][pars_pt->posl[j]][k];
            sR = s_pt->s[h][pars_pt->posr[j]][k];
            rL = dis2rho (s_pt->map[pars_pt->pos[j]]
                          - s_pt->map[pars_pt->posl[j]]);
            rR = dis2rho (s_pt->map[pars_pt->posr[j]]
                          - s_pt->map[pars_pt->pos[j]]);

            for (bL = 0; bL < 2; bL++)
            {
               for (bR = 0; bR < 2; bR++)
               {
                  p1 = exp ((1 - bL) * log (rL) + bL * log (1 - rL) + (1
                                                                       -
                                                                       bR) *
                            log (rR) + bR * log (1 - rR));
                  p0 =
                     exp (bL * log (rL) + (1 - bL) * log (1 - rL) +
                          bR * log (rR) + (1 - bR) * log (1 - rR));
                  p_tab[bL][bR] = (p1 / (p0 + p1));
               }
            }

            for (i = 0; i < n_meis; i++)
            {
               p_sw[i] = p_tab[(sL >> i) & 1][(sR >> i) & 1];
            }

            for (h1 = 0; h1 < pars_pt->n_mc; h1++)
            {
               /*    printf("(h, j, h1) = (%d, %d, %d)\n", h, j, h1); */
               /* Simulate conditional null S */

               /*    printf("(rL, rR) = (%lf, %lf)\n", */
               /*    (s_pt->map[pars_pt->pos[j]] - s_pt->map[pars_pt->posl[j]]), */
               /*    (s_pt->map[pars_pt->posr[j]] - s_pt->map[pars_pt->pos[j]])); */
               /*    printf("(rL, rR) = (%lf, %lf)\n", rL, rR); */

               /* s0 = simulate_linked_s(adg_pt->peds[k].n_nfa, sL, sR, p_tab); */

               /* This might be the most time-consuming line */
               s0 = simulate_binary_vec (n_meis, p_sw);

               /*                 printf("sL, sR, s0\n");
                  print_bits(sL, adg_pt->peds[k].n_nfa);
                  printf("\n");
                  print_bits(sR, adg_pt->peds[k].n_nfa);
                  printf("\n");
                  print_bits(s0, adg_pt->peds[k].n_nfa);
                  printf("\n");
                */
               /* rep_s = rep_iv(s0, &adg_pt->peds[k]); */
               /* red_s = reduced_iv(rep_s, &adg_pt->peds[k]); */
               /* lgw0[h][j][h1] += lgw_tab.lgw[red_s]; */
               /* lgw0[h][j][h1] += lgw_tab.lgw[adg_pt->peds[k].red_s[s0]]; */
               sumphi0[h][j][h1] += vc_tab.sumphi[red_s_pt[s0]];
               phiphi0[h][j][h1] += vc_tab.phiphi[red_s_pt[s0]];
               t0[h][j][h1] += vc_tab.phiz[red_s_pt[s0]];
            }
         }
      }

      free (red_s_pt);
      free_vc_table (&vc_tab);
      free (p_sw);
   }

   printf ("Compute vc scores for observed.\n");
   muz = sumz / nz;
   vz = (zz - sumz * sumz / nz);
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         r = (phiz[h][j] - sumphi[h][j] * muz) / sqrt ((phiphi[h][j]
                                                        -
                                                        sumphi[h][j] *
                                                        sumphi[h][j] / nz) *
                                                       vz);
         pars_pt->t[h][j] = r * sqrt ((nz - 2) / (1 - r * r));
      }
      free (sumphi[h]);
      free (phiz[h]);
      free (phiphi[h]);
   }
   free (sumphi);
   free (phiz);
   free (phiphi);

   printf ("Compute null vc scores.\n");
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            /* t0[h][j][h1] = phiz0[h][j][h1] / sqrt((zz * phiphi0[h][j][h1] - phiz0[h][j][h1]*phiz0[h][j][h1])/nz); */
            r = (t0[h][j][h1] - sumphi0[h][j][h1] * muz) / sqrt ((phiphi0[h][j]
                                                                  [h1] -
                                                                  sumphi0[h][j]
                                                                  [h1] *
                                                                  sumphi0[h][j]
                                                                  [h1] / nz) *
                                                                 vz);
            t0[h][j][h1] = r * sqrt ((nz - 2) / (1 - r * r));
         }
         free (sumphi0[h][j]);
         free (phiphi0[h][j]);
         /* free(phiz0[h][j]); */
      }
      free (sumphi0[h]);
      free (phiphi0[h]);
      /* free(phiz0[h]); */
   }
   free (sumphi0);
   free (phiphi0);
   /* free(phiz0); */

   printf ("Compute p-values.\n");
   /* Ties are not considered here. */
   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         n_sig = 0;
         t = pars_pt->t[h][j];
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            n_sig += (t0[h][j][h1] > t);
         };
         pars_pt->p[h][j] = ((double) n_sig + 1) / ((double) pars_pt->n_mc + 1);
      }
   }

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         free (t0[h][j]);
      }
      free (t0[h]);
   }
   free (t0);
}
