/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 NAME
    evaluate, free_evaluate

 SYNOPSIS

    #include "twoqtl.h"
    double evaluate (int **FGL_a, int **FGL_b, int comp_no);
    void free_evaluate (void);

 DESCRIPTION
    Evaluate log Prob (Trait | FGL_trait) for one component,
    either by enumerating all allelic allocation to FGL or by
    simulating such allocation.

    free_evaluate frees q_trait (obs. quantitative trait).

 INPUT

    FGL_1   observed FGL (founder genome label) at the first locus
    FGL_2   observed FGL at the second locus, NULL if number of QTL is 1
    comp_no 0, 1, ..., n_compnt - 1

    theta, new are reserved for future use, not used right now.

 DETAILS

    If UseMCSummation == FALSE, evaluate does exact sum over trait genes and
    if UseMCSummation == TRUE, it does MC sum with MC length MCRealizations.

    enumerate () goes through all possible alleles of trait genes [using
    next_allele ()], thus is exact but very slow especially for 2 QTL.
    For example, if number of founders are 10, then number of possible
    alleles are 2^20 (== 1,048,576) for each QTL, so this should not be
    used.

    simulate () does Monte Carlo approximation to the exact, using
    importance sampling [sampled each allele with 50% in random_allele()].
    Number of iterations, MCRealizations, needs to be specified by the user
    (better to use more iterations for 2 QTL).

 BUGS
    simulate has not been tested enough yet.
    FGL_1 and FGL_0 are created in mc_twoqtl.c before calling evaluate.

 AUTHOR
    Written by Yun Ju Sung.

 HISTORY
    Documented.                          Apr 15, 2005
    Updated to handle componentwise.     May 19, 2005
    Document revised.                    Jul  2, 2005
    Simulate () tested and a bug fixed.  Jul 14, 2005
    Extended to gene-gene interaction.   Sep 19, 2006

*/

#include "twoqtl.h"

static int obs_size_now;
static int comp_no_now;

double evaluate (int **FGL_1, int **FGL_2, int comp_no)
{
   /* Count the observed FGL.                                                 */
   double *y, log_prob;
   int  *obs_idx_1, n_obs_1;
   int  *obs_idx_2 = NULL, n_obs_2 = 0;

   if ((NQtl == 2) && (FGL_2 == NULL))
      give_up_where (WHERE, ERR_qit);

   comp_no_now = comp_no;
   obs_size_now = obs_size_c[comp_no];

   obs_idx_1 = set_obs_FGL_idx (FGL_1, &n_obs_1);
   if (NQtl == 2)
      obs_idx_2 = set_obs_FGL_idx (FGL_2, &n_obs_2);

   y = (double *) malloc (obs_size_now * sizeof (double));
   alloc_chk (WHERE, y);

   log_prob = 0.0;
   if (UseMCSummation == FALSE)
      log_prob = enumerate (FGL_1, FGL_2, obs_idx_1, obs_idx_2,
                            n_obs_1, n_obs_2, y);
   else
      log_prob = simulate (FGL_1, FGL_2, obs_idx_1, obs_idx_2,
                           n_obs_1, n_obs_2, y);

   free (obs_idx_1);
   if (obs_idx_2 != NULL)
      free (obs_idx_2);
   free (y);

   return log_prob;
}

double enumerate (int **FGL_1, int **FGL_2, int *obs_idx_1, int *obs_idx_2,
                  int n_obs_1, int n_obs_2, double *y)
{
   double foo_v, max_v, value;
   unsigned long int n_geno_1, n_geno_2, i1, i2;

   n_geno_1 = 1;
   n_geno_1 <<= n_obs_1;                /* n_geno = 2^(n_obs_FGL)             */

   n_geno_2 = 1;
   n_geno_2 <<= n_obs_2;

   for (i1 = 0; i1 < n_geno_1; ++i1)
      for (i2 = 0; i2 < n_geno_2; ++i2)
      {
         value = log_prob_jnt (y, obs_idx_1, obs_idx_2, n_obs_1, n_obs_2,
                               FGL_1, FGL_2, i1, i2);
         update_batch (value, (i1 == 0 && i2 == 0), &foo_v, &max_v);
      }

   value = log (foo_v) + max_v;

   if (DebugLevel >= 3)
      fprintf (stderr, "evaluate: %e .....\n", value);

   return value;
}

double simulate (int **FGL_1, int **FGL_2, int *obs_idx_1, int *obs_idx_2,
                 int n_obs_1, int n_obs_2, double *y)
{
   double foo_v, max_v, value;
   unsigned long int i1;
   int   iteration;

   i1 = 0;

   for (iteration = 0; iteration < MCRealizations; ++iteration)
   {
      value = log_prob_jnt (y, obs_idx_1, obs_idx_2, n_obs_1, n_obs_2,
                            FGL_1, FGL_2, i1, i1);
      update_batch (value, (iteration == 0), &foo_v, &max_v);
   }

   value = log (foo_v / MCRealizations) + max_v;

   return value;
}

double log_prob_jnt (double *y, int *obs_idx_1, int *obs_idx_2, int n_obs_1,
                     int n_obs_2, int **FGL_1, int **FGL_2,
                     unsigned long int j, unsigned long int k)
{
   char *allele1, *allele2;
   double log_prob;
   int   n_A;

   memcpy (y, q_trait + obs_ibase[comp_no_now], obs_size_now * sizeof (double));

   if (UseMCSummation == FALSE)
      allele1 = next_allele (n_obs_1, j, &n_A);
   else
      allele1 = random_allele (n_obs_1, sim_freq_1, &n_A);

   log_prob = log_prob_allele (n_A, n_obs_1, Qq[0][1]);

   if (UseMCSummation == TRUE)
      log_prob -= log_prob_allele (n_A, n_obs_1, sim_freq_1);

   if (GxGInteraction == FALSE)
      adj_qtl_effect (y, FGL_1, obs_idx_1, allele1, 1);

   if (DebugLevel == 3)
      fprintf (stderr, "%2lx %2lx %6.4f ", j, k, exp (log_prob));

   if (NQtl == 2)
   {
      if (UseMCSummation == FALSE)
         allele2 = next_allele (n_obs_2, k, &n_A);
      else
         allele2 = random_allele (n_obs_2, sim_freq_2, &n_A);

      log_prob += log_prob_allele (n_A, n_obs_2, Qq[1][1]);

      if (UseMCSummation == TRUE)
         log_prob -= log_prob_allele (n_A, n_obs_2, sim_freq_2);

      if (GxGInteraction == FALSE)
         adj_qtl_effect (y, FGL_2, obs_idx_2, allele2, 2);
      else
         adj_qtl_joint (y, FGL_1, FGL_2, obs_idx_1, obs_idx_2, allele1,
                        allele2);

      free (allele2);
   }

   log_prob += tq_gauss_quant (y, comp_no_now, 0.0, 1.0, 0);
   free (allele1);

   return log_prob;
}

char *next_allele (int n_obs, unsigned long int index, int *n_A_p)
{
   char *allele, current_allele;
   int   n_A;
   int   i1;

   allele = (char *) malloc (n_obs * sizeof (char));
   alloc_chk (WHERE, allele);

   n_A = 0;

   for (i1 = 0; i1 < n_obs; ++i1)
   {
      current_allele = (index & (1 << i1)) != 0;
      n_A += !current_allele;           /* 0 == A  and  1 == a                */
      allele[i1] = current_allele;
/*      fprintf (stderr, "%2d", (int) current_allele); */
   }

   *n_A_p = n_A;

   return allele;
}

char *random_allele (int n_obs, double freq, int *n_A_p)
{
   char *allele, current_allele;
   int   n_A;
   int   i1;

   allele = (char *) malloc (n_obs * sizeof (char));
   alloc_chk (WHERE, allele);

   n_A = 0;

   for (i1 = 0; i1 < n_obs; ++i1)
   {                                    /* Generate A with probability freq.  */
      current_allele = (rans_alt () < freq) ? 0 : 1;
      n_A += !current_allele;           /* 0 == A  and  1 == a                */
      allele[i1] = current_allele;
/*      fprintf (stderr, "%2d", (int) current_allele); */
   }

   *n_A_p = n_A;

   return allele;
}

/* Subtract each QTL effect from the trait.                                   */

void adj_qtl_effect (double *y, int **FGL1, int *obs1, char *allele1,
                     int qtl_no)
{
   char  genotype1;
   int   i1;

   if (DebugLevel == 4)
      fprintf (stderr, ".. adjust qtl effect for qtl %2d ..\n", qtl_no);

   for (i1 = 0; i1 < obs_size_now; ++i1)
   {
      genotype1 = allele1[obs1[FGL1[0][i1] - 1]];
      genotype1 += allele1[obs1[FGL1[1][i1] - 1]];

      if (DebugLevel == 4)
         fprintf (stderr, "%2d %2d %6.4f ", i1, (int) genotype1, y[i1]);

      if (qtl_no == 1)
      {
         y[i1] -= geno_means_1[(int) genotype1];
         if (DebugLevel == 4)
            fprintf (stderr, "%4.2f %6.4f\n",
                     geno_means_1[(int) genotype1], y[i1]);
      }
      else
      {
         y[i1] -= geno_means_2[(int) genotype1];
         if (DebugLevel == 4)
            fprintf (stderr, "%4.2f %6.4f\n",
                     geno_means_2[(int) genotype1], y[i1]);
      }
   }

   return;
}

/* Subtract both QTL effects from the trait (for GxG interaction).            */

void adj_qtl_joint (double *y, int **FGL1, int **FGL2,
                    int *obs1, int *obs2, char *allele1, char *allele2)
{
   char  genotype1, genotype2;
   int   i1;

   if (DebugLevel == 4)
      fprintf (stderr, ".. adjust qtl effect jointly ..\n");

   for (i1 = 0; i1 < obs_size_now; ++i1)
   {
      genotype1 = allele1[obs1[FGL1[0][i1] - 1]];
      genotype1 += allele1[obs1[FGL1[1][i1] - 1]];

      genotype2 = allele2[obs2[FGL2[0][i1] - 1]];
      genotype2 += allele2[obs2[FGL2[1][i1] - 1]];

      if (DebugLevel == 4)
         fprintf (stderr, "%2d %2d %2d %6.4f ", i1, (int) genotype1,
                  (int) genotype2, y[i1]);

      y[i1] -= GxGMeans[(int) genotype1][(int) genotype2];
      if (DebugLevel == 4)
         fprintf (stderr, "%4.2f %6.4f\n",
                  GxGMeans[(int) genotype1][(int) genotype2], y[i1]);
   }

   return;
}

/* Log probability of founder genes.                                          */

double log_prob_allele (int n_A, int n_obs, double P)
{
   double log_prob;

   log_prob = n_A * log (P) + (n_obs - n_A) * log (1.0 - P);

   if (DebugLevel == 5)
      fprintf (stderr, "log_prob_allele %d %d %e %e\n", n_A, n_obs, P,
               log_prob);

   return log_prob;
}

int  *set_obs_FGL_idx (int **FGL, int *size_p)
{
   int  *observed, *obs_FGL_idx;
   int   n_FGL, n_observed, n_obs_FGL;
   int   mat, pat, i1;

   n_FGL = nFGL_c[comp_no_now];
   observed = (int *) malloc (n_FGL * sizeof (int));
   alloc_chk (WHERE, observed);
   memset (observed, 0, n_FGL * sizeof (int));

   n_observed = 0;
   for (i1 = 0; i1 < obs_size_now; ++i1)
   {
      mat = FGL[0][i1] - 1;
      pat = FGL[1][i1] - 1;

      if (observed[mat] == 0)
      {
         observed[mat] = 1;
         ++n_observed;
      }

      if (observed[pat] == 0)
      {
         observed[pat] = 1;
         ++n_observed;
      }

      if (n_observed == n_FGL)
         break;
   }

   obs_FGL_idx = (int *) malloc (n_FGL * sizeof (int));
   alloc_chk (WHERE, obs_FGL_idx);

   n_obs_FGL = 0;
   for (i1 = 0; i1 < n_FGL; ++i1)
      if (observed[i1] == 1)
         obs_FGL_idx[i1] = n_obs_FGL++;
      else
         obs_FGL_idx[i1] = -1;          /* This is out of array index.        */

   if (DebugLevel >= 3)
   {
      fprintf (stderr, "....observed...\n");

      for (i1 = 0; i1 < n_FGL; ++i1)
         fprintf (stderr, "%2d", observed[i1]);
      fprintf (stderr, "  %2d  %2d\n", n_observed, n_FGL);

      for (i1 = 0; i1 < n_FGL; ++i1)
         fprintf (stderr, "%2d", obs_FGL_idx[i1]);
      fprintf (stderr, "  %2d\n", n_obs_FGL);
   }

   if (n_obs_FGL != n_observed)
      give_up_where (WHERE, ERR_qit);

   free (observed);
   *size_p = n_observed;

   return obs_FGL_idx;
}

void set_evaluate (void)
{
   int   index;
   int   i1;

   if (UseMCSummation == TRUE)
   {
      sim_freq_1 = Qq[0][1];
      if (NQtl == 2)
         sim_freq_2 = Qq[1][1];
   }

   if (nFGL_c == NULL)
      nFGL_c = get_nFGL_c ();

   if (q_trait == NULL)
   {
      /* Nq is the number of observed individuals in the pedigree.            */
      q_trait = (double *) malloc (Nq * sizeof (double));
      alloc_chk (WHERE, q_trait);

      index = 0;                        /* q_trait is 0-origin-indexed.       */
      for (i1 = 0; i1 < pedsize; ++i1)
         if (nghd[i1].q_dat != NULL)    /* Is the individual observed?        */
            q_trait[index++] = nghd[i1].q_dat[0];
   }

   return;
}

void free_evaluate (void)
{
   if (q_trait != NULL)
      free (q_trait);

   return;
}
